/*
** Copyright (C) 2003-2006 Teus Benschop.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "backup.h"
#include "generalconfig.h"
#include "projectconfig.h"
#include "gwrappers.h"
#include "date_time_utils.h"
#include "statistics.h"
#include "utilities.h"
#include "projectutils.h"
#include "books.h"
#include "unixwrappers.h"
#include "directories.h"
#include "shell.h"
#include "gtkwrappers.h"
#include "htmlbrowser.h"
#include "subversion.h"


void backup_make_incremental ()
{
  GeneralConfiguration genconfig (0);
  ProjectConfiguration projectconfig (genconfig.project());
  backup_make (genconfig.project(), false, projectconfig.backup_incremental_last_time());
  projectconfig.backup_incremental_last_time_set (date_time_seconds_get_current ());
}


void backup_make_flexible ()
{
  GeneralConfiguration genconfig (0);
  ProjectConfiguration projectconfig (genconfig.project());
  backup_make (genconfig.project(),
               projectconfig.backup_full (),
               projectconfig.backup_incremental_last_time());
  if (!projectconfig.backup_full())
    projectconfig.backup_incremental_last_time_set (date_time_seconds_get_current ());
}


void backup_make (const ustring& project, bool full, int timefrom)
{
  // Working directory and zipped filename.
  ustring workingdirectory = gw_build_filename (directories_get_temp (), "backup");
  unix_mkdir (workingdirectory);
  int currentseconds = date_time_seconds_get_current ();
  int currentdays = date_time_julian_day_get_current ();
  guint currentyear, currentmonth, currentday;
  int currenthour, currentminute, currentsecond;
  date_time_normal_get_year_month_day (currentdays, currentyear, currentmonth, currentday);
  date_time_normal_get_hour_minute_second (currentseconds, currenthour, currentminute, currentsecond);
  ustring currentsuffix = "-"
                        + convert_to_string (currentyear) + "-" 
                        + convert_to_string (currentmonth) + "-"
                        + convert_to_string (currentday) + "-"
                        + convert_to_string (currenthour) + "-"
                        + convert_to_string (currentminute) + "-"
                        + convert_to_string (currentsecond);  
  ustring filename;
  if (full) {
    filename = "full-backup" + currentsuffix;
  } else {
    guint fromyear, frommonth, fromday;
    int fromhour, fromminute, fromsecond;
    date_time_normal_get_year_month_day (date_time_seconds_to_julian (timefrom), fromyear, frommonth, fromday);
    date_time_normal_get_hour_minute_second (timefrom, fromhour, fromminute, fromsecond);
    ustring fromsuffix = "-from-"
                       + convert_to_string (fromyear) + "-" 
                       + convert_to_string (frommonth) + "-"
                       + convert_to_string (fromday) + "-"
                       + convert_to_string (fromhour) + "-"
                       + convert_to_string (fromminute) + "-"
                       + convert_to_string (fromsecond);  
    filename = "incremental-backup" + fromsuffix + "-to" + currentsuffix;
  }
  filename.append (".zip");
  filename = gw_build_filename (workingdirectory, filename);
  
  if (full) {
    // Handle full backup first.
    GwSpawn spawn ("bibledit");
    spawn.arg ("--export-project-usfm");
    spawn.arg ("--gui");
    spawn.arg ("--project");
    spawn.arg (project);
    spawn.arg ("--location");
    spawn.arg (filename);
    spawn.arg ("--zip");
    spawn.run ();
  } else {
    // Handle incremental backups.
    // Get the books/chapters changed since a given time.
    // Get them from subversion, because subversion also records changes
    // made by others. Locally produced tracking mechanisms would not so easily 
    // provide that information.
    vector <unsigned int> books;
    vector <unsigned int> chapters;
    subversion_get_chapters_changed_since (project, timefrom, books, chapters);
    if (books.empty()) {
      gtkw_dialog_info (NULL, "There was nothing to backup");
      return;
    }
    quick_sort (books, chapters, 0, books.size());
    // Storage for the data of the chapters in each book.
    map <unsigned int, vector<ustring> > lines;
    // Always include chapter 0 as that has the id line.
    set <unsigned int> books_chapter_zero_done;
    for (unsigned int i = 0; i < books.size(); i++) {
      if (books_chapter_zero_done.find (books[i]) == books_chapter_zero_done.end()) {
        vector <ustring> data = project_retrieve_chapter (project, books[i], 0);
        for (unsigned int i2 = 0; i2 < data.size(); i2++) {
          lines [books[i]].push_back (data[i2]);
        }
        books_chapter_zero_done.insert (books[i]);
      }
    }
    // Go through the books/chapters with changes, and add the text to the map.
    // Skip chapter 0. Has been done already.
    for (unsigned int i = 0; i < books.size(); i++) {
      if (chapters[i] != 0) {
        vector <ustring> data = project_retrieve_chapter (project, books[i], chapters[i]);
        for (unsigned int i2 = 0; i2 < data.size(); i2++) {
          lines [books[i]].push_back (data[i2]);
        }
      }
    }
    // Go through all books, see which ones have data, save to file.
    vector <unsigned int> all_books = books_type_to_ids (btUnknown);
    for (unsigned int i = 0; i < all_books.size(); i++) {
      if (!lines[books[i]].empty()) {
        ustring usfmfilename = gw_build_filename (workingdirectory, books_id_to_english (books[i]) + ".usfm");
        write_lines (usfmfilename, lines[books[i]]);
      }
    }
    // Zip them.
    ustring command = "cd" + shell_quote_space (workingdirectory) + "; zip -r " 
                    + gw_path_get_basename (filename) + " *.usfm; rm *.usfm"; // todo fix on Windows.
    system (command.c_str());
  }

  // Open web page with information: backup_is_ready.html.
  ustring filesize = convert_to_string (file_get_size (filename));
  ReadText rt (gw_build_filename (directories_get_package_data(), "backup_is_ready.html"), true, false);
  for (unsigned int i = 0; i < rt.lines.size(); i++) {
    replace_text (rt.lines[i], "100", filesize);
    replace_text (rt.lines[i], "location", filename);    
    replace_text (rt.lines[i], "tempdir", workingdirectory);    
  }
  ustring htmlfile = gw_build_filename (workingdirectory, "backup.html");
  write_lines (htmlfile, rt.lines);
  unix_cp (gw_build_filename (directories_get_package_data(), "bibledit.css"), gw_build_filename (workingdirectory, "bibledit.css"));
  htmlbrowser (htmlfile, false);  
}
