/*
** Copyright (C) 2003-2006 Teus Benschop.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "utilities.h"
#include <glib.h>
#include "constants.h"
#include "projectutils.h"
#include "directories.h"
#include "gwrappers.h"
#include "shell.h"
#include "sqlite3.h"
#include "books.h"
#include "vacuum.h"
#include "bible.h"
#include "progresswindow.h"
#include "import.h"
#include "unixwrappers.h"
#include "statistics.h"
#include "subversion.h"
#include "projectconfig.h"
#include "dialoglistview.h"
#include <gdk/gdkkeysyms.h>


void project_store_sanitize_line (ustring& line)
/* 
Joins lines, without starting usfm, to a previous one. 
Makes changes in-place.
*/
{
  // Look for a newline.
  size_t position = line.find ("\n");
  while (position != string::npos) {
    position++;
    if (position >= line.size())
      break;
    // If found, and the next character is not a backslash (\), then put a space 
    // instead of the newline.
    if (line.substr (position, 1) != "\\") {
      line.replace (position - 1, 1, " ");
    }
    // Keep going till through.
    position = line.find ("\n", position);
  }
}


void project_store_chapter_internal (
       const ustring& project, unsigned int book, unsigned int chapter, 
       CategorizeChapterVerse& ccv, unsigned int first, unsigned int last,
       unsigned int timestamp)
{
  // Store statistics.
  statistics_record_store_chapter (project, book, chapter, ccv, first, last);
  // Store data.
  try
  {
    ustring chapterdirectory = project_data_directory_chapter (project, book, chapter);
    if (!g_file_test (chapterdirectory.c_str(), G_FILE_TEST_IS_DIR))
      unix_mkdir (chapterdirectory);
    ustring filename = project_data_filename_chapter (project, book, chapter, false);
    vector <ustring> data;
    for (unsigned int i = first; i < last; i++) {
      ustring text = ccv.line[i];
      project_store_sanitize_line (text);
      data.push_back (text);
    }    
    write_lines (filename, data);
  }
  catch (exception & ex)
  {
    gw_critical (ex.what ());
  }
  // Timestamp.
  project_update_book (project, book, timestamp);
  // Subversion.
  subversion_store_chapter (project, book, chapter);
}


void projects_initial_check (bool gui)
// Does some initial checks on the projects we have.
// Upgrades projects of an older version to the current version.
{
}


vector<ustring> projects_get_all ()
// Gets sorted list of all projects.
{
  ReadDirectories rd (directories_get_projects (), "", "");
  sort (rd.directories.begin(), rd.directories.end());
  return rd.directories;
}


ustring project_filename (const ustring& project)
// Returns the filename of the database that keeps synchronization data about 
// the project.
{
  return gw_build_filename (directories_get_projects (), project, "data.sql2");
}


ustring project_data_directory_project (const ustring& project)
// Returns the data directory for the project, e.g.:
// ~/.bibledit/projects/testproject/data
{
  ustring directory = gw_build_filename (directories_get_projects(), project, "data");
  return directory;
}


ustring project_data_directory_book (const ustring& project, unsigned int book)
// Returns the data directory for the book, e.g.:
// ~/.bibledit/projects/testproject/data/Genesis
{
  ustring directory = gw_build_filename (project_data_directory_project (project), books_id_to_english (book));
  return directory;
}


ustring project_data_directory_chapter (const ustring& project, unsigned int book, unsigned int chapter)
// Returns the data directory for the chapter, e.g.:
// ~/.bibledit/projects/testproject/data/Genesis/1
{
  ustring directory = gw_build_filename (project_data_directory_book (project, book), convert_to_string (chapter));
  return directory;
}


ustring project_data_filename_chapter (const ustring& project, unsigned int book, unsigned int chapter, bool checkexists)
// Returns the data filename of the chapter, e.g.:
// ~/.bibledit/projects/testproject/data/Genesis/1/data
{
  ustring filename = gw_build_filename (project_data_directory_chapter (project, book, chapter), "data");
  if (checkexists) {
    if (!g_file_test (filename.c_str(), G_FILE_TEST_IS_REGULAR))
      throw runtime_error (books_id_to_english (book) + " " + convert_to_string (chapter) + " does not exist");
  }
  return filename;
}


bool project_exists (const ustring& project)
{
  return g_file_test (gw_build_filename (directories_get_projects(), project).c_str(), G_FILE_TEST_IS_DIR);
}


void project_create (const ustring& project)
{
  // Create project directory, and directory for folded data.
  unix_mkdir (project_data_directory_project (project));
  // Create synchronization and index database.  
  sqlite3 *db;
  int rc;
  char *error = NULL;
  try
  {
    rc = sqlite3_open (project_filename (project).c_str (), &db);
    if (rc) throw runtime_error (sqlite3_errmsg(db));
    sqlite3_busy_timeout (db, 1000);
    char * sql;
    // Create the synchronization table.
    // It stores the state of the synchronization between internal database and external USFM files.
    sql = g_strdup_printf ("create table sync (book text, updated integer, filename text, filetime integer, filesize integer);");
    rc = sqlite3_exec (db, sql, NULL, NULL, &error);
    g_free (sql);
    if (rc) {
      throw runtime_error (sqlite3_errmsg(db));
    }
    // Create the index table.
    // Information to keep the index up-to-date.
    sql = g_strdup_printf ("create table idx (timestamp integer);");
    rc = sqlite3_exec (db, sql, NULL, NULL, &error);
    g_free (sql);
    if (rc) {
      throw runtime_error (sqlite3_errmsg(db));
    }
  }
  catch (exception & ex)
  {
    gw_critical (ex.what ());
  }
  sqlite3_close (db);
  // Create statistics system.
  statistics_initial_check_project (project, false);
  // Create subversion repository.
  subversion_initial_check_project (project);
}


void project_delete (const ustring& project)
// Deletes a project and associated data.
{
  // No project: bail out.
  if (project.empty()) return;
  // Remove repositories.
  subversion_remove_project (project);
  // Delete the whole project, including the statistics database.
  ustring directory = gw_build_filename (directories_get_projects (), project);
  unix_rmdir (directory);
}


void project_copy (const ustring& project, const ustring& newproject)
// Copies "project" to "newproject.
{
  ustring olddirectory = gw_build_filename (directories_get_projects (), project);
  ustring newdirectory = gw_build_filename (directories_get_projects (), newproject);
  unix_cp_r (olddirectory, newdirectory);
  // Statistics have been copied too. Do initial check.
  statistics_initial_check_project (newproject, false);
  // Use of a remote repository is switched off so as not to mess with our server. 
  subversion_revert_to_internal_repository (newproject);
}


void project_move (const ustring& project, const ustring& newproject)
{
  // Move subversion project.
  subversion_move_project (project, newproject);
  // This moves the whole project, including the statistics database.
  ustring oldname = (gw_build_filename (directories_get_projects (), project));
  ustring newname = (gw_build_filename (directories_get_projects (), newproject));
  unix_mv (oldname, newname);
}


void project_store_book (const ustring& project, unsigned int book, CategorizeChapterVerse& ccv)
{
  // If nothing to save, bail out.
  if (ccv.chapter.empty()) return;
  // Timestamp.
  unsigned int timestamp = time (0);
  // Make pointers to each chapter.
  vector <unsigned int> chapters;
  vector <unsigned int> firsts;
  vector <unsigned int> lasts;
  unsigned int currentchapter = ccv.chapter[0];
  unsigned int first = 0;
  for (unsigned int i = 0; i < ccv.chapter.size(); i++) {
    if (ccv.chapter[i] != currentchapter) {
      chapters.push_back (currentchapter);
      currentchapter = ccv.chapter[i];
      firsts.push_back (first);
      first = i;
      lasts.push_back (i);
    }
  }
  chapters.push_back (currentchapter);
  firsts.push_back (first);
  lasts.push_back (ccv.chapter.size());
  // Store each chapter.
  for (unsigned int i = 0; i < chapters.size(); i++) {
    project_store_chapter_internal (project, book, chapters[i], ccv, firsts[i], lasts[i], timestamp);
  }  
  // Store data about book.
  project_update_book (project, book, timestamp);
}


void project_remove_book (const ustring& project, unsigned int book)
{
  // Store statistics.
  statistics_record_remove_book (project, book);
  // Repository update
  subversion_remove_book (project, book);
  // Remove book directory.
  unix_rmdir (project_data_directory_book (project, book));
  // Remove synchronization information.
  sqlite3 *db;
  int rc;
  char *error = NULL;
  try
  {
    rc = sqlite3_open (project_filename (project).c_str (), &db);
    if (rc) throw runtime_error (sqlite3_errmsg(db));
    sqlite3_busy_timeout (db, 1000);
    // Remove possible link.
    char * sql;
    sql = g_strdup_printf ("delete from sync where book = '%s';", books_id_to_english (book).c_str());
    rc = sqlite3_exec (db, sql, NULL, NULL, &error);
    g_free (sql);
    if (rc) {
      throw runtime_error (sqlite3_errmsg(db));
    }
  }
  catch (exception & ex)
  {
    gw_critical (ex.what ());
  }
  // Close connection.
  sqlite3_close (db);
}


void project_store_chapter (const ustring& project, unsigned int book, CategorizeChapterVerse& ccv)
{
  // If there's nothing to save, bail out.
  if (ccv.chapter.size() == 0)
    return;
  // Do the actual storing.
  unsigned int chapter = ccv.chapter[0];
  unsigned int timestamp = time (0);
  project_store_chapter_internal (project, book, chapter, ccv, 0, ccv.chapter.size(), timestamp);
  project_update_book (project, book, timestamp);
}


void project_remove_chapter (const ustring& project, unsigned int book, unsigned int chapter)
{
  // Statistics.
  statistics_record_remove_chapter (project, book, chapter);
  // Actual removal.
  ustring directory = project_data_directory_chapter (project, book, chapter);
  unix_rmdir (directory);
  // Update book.
  project_update_book (project, book, time (0));
  subversion_remove_chapter (project, book, chapter);
}


void project_store_verse (const ustring& project, unsigned int book, unsigned int chapter, const ustring& verse, const ustring& data)
{
  // If there's nothing to save, bail out.
  if (data.empty())
    return;
  // Statistics.
  statistics_record_store_verse (project, book, chapter, verse, data);
  // Retrieve current data, categorize it, store new verse at right place, store it.
  vector <ustring> lines = project_retrieve_chapter (project, book, chapter);
  CategorizeChapterVerse ccv (lines);
  for (unsigned int i = 0; i < ccv.verse.size(); i++) {
    if (ccv.verse[i] == verse) {
      ustring text (data);
      project_store_sanitize_line (text);
      ccv.line[i] = text;
    }
  }
  unsigned int timestamp = time (0);
  project_store_chapter_internal (project, book, chapter, ccv, 0, ccv.chapter.size(), timestamp);
  // Update book.
  project_update_book (project, book, timestamp);
  subversion_store_chapter (project, book, chapter);
}


void project_update_book (const ustring& project, unsigned int book, unsigned int timestamp)
// Sets flags/variables that indicate this book has been updated.
{
  // Some variables.  
  sqlite3 *db;
  int rc;
  char *error = NULL;
  try
  {
    // Connect to the database.
    rc = sqlite3_open (project_filename (project).c_str (), &db);
    if (rc) throw runtime_error (sqlite3_errmsg(db));
    sqlite3_busy_timeout (db, 1000);
    // Variable.
    char * sql;
    // Delete the timestamp.
    sql = g_strdup_printf ("delete from idx;");
    rc = sqlite3_exec (db, sql, NULL, NULL, &error);
    g_free (sql);
    if (rc) {
      throw runtime_error (sqlite3_errmsg(db));
    }
    // Store timestamp
    sql = g_strdup_printf ("insert into idx values (%d)", timestamp);
    rc = sqlite3_exec (db, sql, NULL, NULL, &error);
    g_free (sql);
    if (rc) {
      throw runtime_error (sqlite3_errmsg(db));
    }
    // Set updated flag for synchronization, if book given.
    if (book != 0) {
      sql = g_strdup_printf ("update sync set updated = %d where book = '%s';", 1, books_id_to_english (book).c_str());
      rc = sqlite3_exec (db, sql, NULL, NULL, &error);
      g_free (sql);
      if (rc) {
        throw runtime_error (sqlite3_errmsg(db));
      }    
    }
  }
  catch (exception & ex)
  {
    gw_critical (ex.what ());
  }
  // Close connection.
  sqlite3_close (db);
}


ustring project_retrieve_verse_extended (const ustring& project, unsigned int book, unsigned int chapter, const ustring& verse)
/*
This function will be called when the exact verse was not found in the text.
E.g. it requested verse "1-2", and the text happens to have verse "1" and
verse "2", but not the combined verse. This extended function handles these
and similar cases.
This is roughly the strategy to be followed to get the right verse.
- The requested verse may have the a or b suffix, e.g. verse 4a.
- Therefore use a system of "half" verses.
- Create a vector with all the half verses requested.
- Create sets with all the half verses that are in a particular verse, e.g.
  v 1-2 will contain 1a and 1b, and v 3-5 contains 3a, 3b, 4a, 4b, 5a, 5b.
- Take any set that contains any of the requested half verses and get the 
  text(s) of the verse(s).
*/
{
  // Holds the verse text we will retrieve.
  ustring line;
  
  // Get the requested half-verses.
  vector<int> requested_verses = verses_encode (verse);

  // Get all verses in this chapter.
  vector <ustring> all_verses = project_get_verses (project, book, chapter);

  // Get the sets with the half-verses of each available verse in this chapter.
  vector< set<int> > available_verses_sets;
  for (unsigned int i = 0; i < all_verses.size (); i++) {
    vector<int> verses_v = verses_encode (all_verses[i]);
    set<int> verses_s (verses_v.begin(), verses_v.end());
    available_verses_sets.push_back (verses_s);
  }

  // Retrieve any verse that relates to the verse we look for.
  for (unsigned int a = 0; a < available_verses_sets.size(); a++) {
    for (unsigned int r = 0; r < requested_verses.size(); r++) {
      if (available_verses_sets[a].find (requested_verses[r]) != available_verses_sets[a].end()) {
        if (!line.empty ()) 
          line.append ("\n");
        line.append (project_retrieve_verse (project, book, chapter, all_verses[a]));
      }
    }
  }

  // Return whatever we got.
  return line;   
}


ustring project_retrieve_verse (const ustring& project, unsigned int book, unsigned int chapter, const ustring& verse)
{
  // Retrieve the verse.
  ustring line;
  vector <ustring> lines = project_retrieve_chapter (project, book, chapter);
  CategorizeChapterVerse ccv (lines);
  for (unsigned int i = 0; i < ccv.verse.size(); i++) {
    if (ccv.verse[i] == verse) {
      line = ccv.line[i];
    }
  }
  // If the verse was not found, use extended retrieval mechanism to get it.
  if (line.empty()) {
    line = project_retrieve_verse_extended (project, book, chapter, verse);
  }
  // Return the data.
  return line;
}


vector<ustring> project_retrieve_chapter (const ustring& project, unsigned int book, unsigned int chapter)
// Get the chapter from disk.
{
  // Get the data.
  ustring filename = project_data_filename_chapter (project, book, chapter, false);
  ReadText rt (filename, true);
  // If the data shows that there is a conflict, solve it, then reload.
  bool solve_conflict = false;
  for (unsigned int i = 0; i < rt.lines.size(); i++) {
    if (rt.lines[i] == subversion_mine_conflict_marker ()) {
      solve_conflict = true;
    }
  }
  if (solve_conflict) {
    subversion_resolve_conflict_chapter (project, book, chapter);
    ReadText rt2 (filename, true);
    return rt2.lines;
  }
  // Return the data.
  return rt.lines;
}


vector<ustring> project_retrieve_book (const ustring& project, unsigned int book)
// Retrieves the text of the whole book.
{
  // Storage for the data.
  vector<ustring> lines;
  // All chapters.
  vector<unsigned int> chapters = project_get_chapters (project, book);
  // Get text of each chapter.
  for (unsigned int i = 0; i < chapters.size(); i++) {
    vector<ustring> ln = project_retrieve_chapter (project, book, chapters[i]);
    for (unsigned int i2 = 0; i2 < ln.size(); i2++) {
      lines.push_back (ln[i2]);
    }
  }
  // Return data.
  return lines;
}


unsigned int project_retrieve_timestamp (const ustring& project)
{
  // Holds the retrieved timestamp.
  unsigned int timestamp = time (0);
  // Database variables.  
  sqlite3 *db;
  int rc;
  char *error = NULL;
  try
  {
    // Open database.
    rc = sqlite3_open(project_filename (project).c_str (), &db);
    if (rc) throw runtime_error (sqlite3_errmsg(db));
    sqlite3_busy_timeout (db, 1000);
    // Read the timestamp
    SqliteReader sqlitereader (0);
    char * sql;
    sql = g_strdup_printf ("select timestamp from idx;");
    rc = sqlite3_exec(db, sql, sqlitereader.callback, &sqlitereader, &error);
    g_free (sql);
    if (rc != SQLITE_OK) {
      throw runtime_error (error);
    }
    if ((sqlitereader.ustring0.size() > 0)) {
      timestamp = convert_to_int (sqlitereader.ustring0[0]);
    }
  }
  catch (exception & ex)
  {
    gw_critical (ex.what ());
  }
  // Close connection.  
  sqlite3_close (db);
  // Return the timestamp.
  return timestamp;
}


void project_link_book (const ustring& project, unsigned int book, const ustring& file)
// This does the administration for linking "book" to "file".
{
  // Some variables.  
  sqlite3 *db;
  int rc;
  char *error = NULL;
  try
  {
    // Connect to the database.
    rc = sqlite3_open (project_filename (project).c_str (), &db);
    if (rc) throw runtime_error (sqlite3_errmsg(db));
    sqlite3_busy_timeout (db, 1000);
    // Variable.
    char * sql;
    // Delete any existing link for this book.
    sql = g_strdup_printf ("delete from sync where book = '%s';", books_id_to_english (book).c_str()); 
    rc = sqlite3_exec (db, sql, NULL, NULL, &error);
    if (rc) {
      throw runtime_error (sqlite3_errmsg(db));
    }
    // Store new link.
    unsigned int filetime = file_get_modification_time (file);
    unsigned int filesize = file_get_size (file);
    sql = g_strdup_printf ("insert into sync values ('%s', %d, '%s', %d, %d)", 
                           books_id_to_english (book).c_str(), 0, file.c_str(), filetime, filesize);
    rc = sqlite3_exec (db, sql, NULL, NULL, &error);
    if (rc) {
      throw runtime_error (sqlite3_errmsg(db));
    }
    // Free memory.
    g_free (sql);
  }
  catch (exception & ex)
  {
    gw_critical (ex.what ());
  }
  // Close connection.
  sqlite3_close (db);
}


ProjectLinksGet::ProjectLinksGet (const ustring& project)
// This retrieves data for all linked books.
{
  sqlite3 *db;
  int rc;
  char *error = NULL;
  try
  {
    rc = sqlite3_open (project_filename (project).c_str (), &db);
    if (rc) throw runtime_error (sqlite3_errmsg(db));
    sqlite3_busy_timeout (db, 1000);
    SqliteReader sqlitereader (0);
    char * sql;
    sql = g_strdup_printf ("select book, updated, filename, filetime, filesize from sync;");
    rc = sqlite3_exec(db, sql, sqlitereader.callback, &sqlitereader, &error);
    g_free (sql);
    if (rc) {
      throw runtime_error (sqlite3_errmsg(db));
    }
    if (sqlitereader.ustring0.size () > 0) {
      for (unsigned int i = 0; i < sqlitereader.ustring0.size(); i++) {
        books.push_back (books_english_to_id (sqlitereader.ustring0[i]));
        updateds.push_back (convert_to_bool (sqlitereader.ustring1[i]));
        filenames.push_back (sqlitereader.ustring2[i]);
        filetimes.push_back (convert_to_int (sqlitereader.ustring3[i]));
        filesizes.push_back (convert_to_int (sqlitereader.ustring4[i]));      
      }
    }
  }
  catch (exception & ex)
  {
    gw_critical (ex.what ());
  }
  sqlite3_close (db);
}


vector<unsigned int> project_get_books (const ustring& project)
// Gives all books in this project.
// Reorders the book according to settings made by user.
{
  vector <unsigned int> books;
  ReadDirectories rd (project_data_directory_project (project), "", "");
  for (unsigned int i = 0; i < rd.directories.size(); i++) {
    unsigned int book = books_english_to_id (rd.directories[i]);
    if (book) books.push_back (book);
  }
  books_order (project, books);
  return books;
}


vector<unsigned int> project_get_chapters (const ustring& project, unsigned int book)
// Gives all chapters in a book of a project.
{
  vector <unsigned int> chapters;
  ReadDirectories rd (project_data_directory_book (project, book), "", "");
  for (unsigned int i = 0; i < rd.directories.size(); i++) {
    // It also reads the .svn directory. Filter that out.
    ustring uchapter = number_in_string (rd.directories[i]);
    if (!uchapter.empty()) {
      unsigned int chapter = convert_to_int (rd.directories[i]);
      chapters.push_back (chapter);
    }
  }
  sort (chapters.begin(), chapters.end());
  return chapters;
}


vector<ustring> project_get_verses (const ustring& project, unsigned int book, unsigned int chapter)
// Returns the verses in a chapter.
{
  vector<ustring> verses;
  vector<ustring> lines = project_retrieve_chapter (project, book, chapter);
  CategorizeChapterVerse ccv (lines);
  verses = ccv.verse;
  return verses;
}


bool project_book_exists (const ustring& project, unsigned int book)
{
  ustring directory = project_data_directory_book (project, book);
  return g_file_test (directory.c_str(), G_FILE_TEST_IS_DIR);  
}


void project_vacuum (const ustring& project, unsigned int starttime)
{
  vacuum_database (project_filename (project), starttime);
}


bool project_select (ustring& newproject)
// Shows the select project dialog. 
// newproject: gets the new project if one was chosen.
// Returns true if a project was chosen.
{
  // Divide the projects into editable and non-editable ones.
  vector <ustring> editable_projects;
  vector <ustring> non_editable_projects;
  {
    vector<ustring> projects = projects_get_all ();
    for (unsigned int i = 0; i < projects.size(); i++) {
      ProjectConfiguration projectconfig (projects[i]);
      if (projectconfig.editable()) editable_projects.push_back (projects[i]);
      else non_editable_projects.push_back (projects[i]);
    }
  }
  // Build/run dialog.    
  ustring focusedproject = newproject;
  if (focusedproject.empty()) {
    GeneralConfiguration genconfig (0);
    focusedproject = genconfig.project();
  }
  ListviewDialog dialog ("Open project", editable_projects, focusedproject, true, NULL);
  dialog.two_pages ("Editable", "Non-editable", non_editable_projects, true);
  dialog.tab_switcher (GDK_O, GDK_CONTROL_MASK, "Ctrl-O", "to display the editable projects", "to display the non-editable projects");
  if (dialog.run () == GTK_RESPONSE_OK) {
    newproject = dialog.focus;
    return true;
  } else {
    return false;
  }
}
